html, body {
  height: 100%;
  overflow: hidden;
  display: flex;
  flex-direction: column;
}
svg {
  width: 100%;
  height: 100%;
  float: left;
  border-bottom: solid 1px #ccc;
  border-right: solid 1px #ccc;
  margin-right: -1px;
  margin-bottom: -1px;
}

#editor_div { 
  width: 100%;
  padding-bottom: 8px;
  position: relative;
  flex: 1;
}
.active-wiring {
  fill-opacity: 0.2 !important;
}
.popup {
    position: absolute;
    background-color: #fff;
    /*width: 200px;*/
    border: 1px #ccc solid;
    border-radius: 6px;
    box-shadow: #333 2px 2px 4px;
    font-family: arial, helvetica, sans-serif;
    margin: 5px 5px 2px 2px;
}
.popup ul {
    padding: 5px;
    list-style-type: none;
    margin: 0;
}

.popup ul li {
    margin-top: 2px;
    margin-bottom: 2px;
    margin-right: 0px;
    margin-left: 0px;
    cursor: pointer;
}
.popup ul li:hover {
    /*background-color: LightGray;*/
}

.popup hr {
    margin: 0;
}

.input.label, .output.label {
    display: none;
}

/* styles for dataflow editor */

g.exposed-terminals rect, g.module rect {
  stroke-width: 2px;
  stroke: blue;
  cursor: move;
  fill: white;
}

svg.editor g.module rect {
  stroke-width: 2px;
  stroke: blue;
}

svg.editor .highlight rect.title {
  stroke-width: 5px;
}

rect.unexplode {
  cursor: none;
  fill: none;
}

g.module rect.title {
  fill-opacity: 0;
}

g.module rect.border.embedded-indicator {
  fill: none;
  stroke: blue;
  stroke-width: 2px;
}

g.module rect.terminal, g.exposed-terminals rect.terminal { 
  cursor: crosshair;
}

rect.terminal.output, rect.terminal.exposed.input {
  fill: #88FFFF;
}

rect.terminal.input, rect.terminal.exposed.output {
  fill:  #88FF88;
}

path.wire, path.exposed-wire {
  fill: none;
  stroke-width: 9px;
  stroke: red;
  cursor: pointer;
}

path.exposed-wire {
  stroke: blue;
  stroke-opacity: 0.4;
}

polygon.terminal.state {
  display: none;
}
